# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals

class MultiBase(object):
    """Represents a base where not every digit has the same possible values.

    The ``bases`` parameter must be a sequence of strings, where each item in
    the sequence represents one digit, in order from most significant digit to
    least significant digit. Each character in each string represents one
    possible value for the corresponding digit, in order from the lowest to
    highest value for that digit.

    The ``max_encodable_value`` attribute is the largest integer that can be
    encoded with this base.
    """

    def __init__(self, bases):
        self.bases = bases

    def encode(self, n):
        """Encode an integer to a string, using this base.

        The ``n`` parameter must be an integer. Returns the encoded string, or
        raises ``ValueError`` if ``n`` is greater than the largest encodable
        integer.
        """

        if n > self.max_encodable_value:
            raise ValueError(
                '%d is greater than the largest encodable integer (%d)' % (
                    n, self.max_encodable_value))
        ret = []
        for base in reversed(self.bases):
            n, d = divmod(n, len(base))
            ret.append(base[d])
        ret.reverse()
        return ''.join(ret)

    def decode(self, x):
        """Decode a string to an integer, using this base.

        The ``x`` parameter must be a string as long as the ``bases`` sequence.
        Returns the decoded integer or raises ``ValueError`` if the length of
        ``x`` is not equal to the length of ``bases`` or any of the characters
        in ``x`` aren't valid digits for their position.
        """

        if len(x) != len(self.bases):
            raise ValueError(
                "the length of %r (%d) doesn't match the number of bases (%d)" % (
                    x, len(x), len(self.bases)))
        ret = 0
        for base, d in zip(self.bases, x):
            ret = (ret * len(base)) + base.index(d)
        return ret

    @property
    def max_encodable_value(self):
        ret = 1
        for base in self.bases:
            ret *= len(base)
        return ret - 1



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import pytest

from passacre import generator


@pytest.mark.skipif("sys.version_info < (3,)")
def test_patched_skein():
    import skein
    r1 = generator._patch_skein_random(skein.Random(b'123'))
    s1 = r1.read(2)
    r1.getrandbits(32)
    s2 = r1.read(2)
    r2 = skein.Random(b'123')
    assert s1 + s2 != r2.read(4)
    with pytest.raises(ValueError):
        r1.getrandbits(0)
    with pytest.raises(ValueError):
        r1.getrandbits(-1)

def test_invalid_prng_method():
    options = {'method': 'invalid', 'iterations': 12}
    with pytest.raises(ValueError):
        generator.build_prng(None, 'passacre', 'example.com', options)



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals, print_function

from passacre.schema import multibase_of_schema
from passacre.util import nested_set, jdumps
from passacre import generator

import collections
import json
import os
import sys


class ConfigBase(object):
    def __init__(self):
        self.defaults = {
            'method': 'keccak',
            'iterations': 1000,
        }
        self.site_hashing = {
            'enabled': True,
        }
        self.global_config = {}
        self.words = None
        self.word_list_file = None

    def load_words_file(self, path):
        if path is None:
            return
        self.word_list_file = path
        try:
            infile = open(os.path.expanduser(path))
        except IOError as e:
            print("warning: couldn't open %r: %s" % (path, e), file=sys.stderr)
        else:
            with infile:
                self.words = [word.strip() for word in infile]

    def fill_out_config(self, config):
        config['multibase'] = multibase_of_schema(config['schema'], self.words)
        config['iterations'] = (
            config.get('iterations', 1000) + config.get('increment', 0))

    def set_defaults(self, defaults):
        self.defaults.update(defaults)
        self.site_hashing.update(
            (k, v) for k, v in self.defaults.items()
            if k in ('method', 'iterations'))
        self.fill_out_config(self.defaults)

    def _get_hashed_site(self, site, password):
        hashed_site = generator.hash_site(password, site, self.site_hashing)
        return self._get_site(hashed_site)

    def get_site(self, site, password=None):
        if self.site_hashing['enabled'] == 'always' and site != 'default':
            config = self._get_hashed_site(site, password)
        else:
            config = self._get_site(site)
        if config is None and password:
            config = self._get_hashed_site(site, password)
        if config is None:
            config = self.defaults
        return config

    def generate_for_site(self, username, password, site):
        config = self.get_site(site, password)
        return generator.generate(username, password, site, config)


class YAMLConfig(ConfigBase):
    def read(self, infile):
        "Load site configuration from a YAML file object."
        import yaml
        parsed = yaml.load(infile)
        sites = parsed.pop('sites', {})
        self.set_defaults(sites.get('default', {}))
        self.load_words_file(parsed.pop('words-file', None))

        self.sites = {}
        for site, additional_config in sites.items():
            site_config = self.sites[site] = self.defaults.copy()
            site_config.update(additional_config)
            self.fill_out_config(site_config)

        self.site_hashing.update(parsed.pop('site-hashing', {}))
        self.global_config = parsed

    def _get_site(self, site, password=None):
        return self.sites.get(site)

    def get_all_sites(self):
        return self.sites

    def _no_config_modification(self, *a, **kw):
        raise NotImplementedError("YAMLConfig doesn't implement configuration modification.")

    add_site = remove_site = set_site_schema = _no_config_modification
    add_schema = remove_schema = set_schema_value = set_schema_name = _no_config_modification

    def get_all_schemata(self, *a, **kw):
        raise NotImplementedError("YAMLConfig doesn't have a way to list schemata.")

    get_schema = get_all_schemata


def maybe_json(val):
    try:
        return json.loads(val)
    except ValueError:
        return val

def maybe_json_dict(pairs):
    return dict((k, maybe_json(v)) for k, v in pairs)

class SqliteConfig(ConfigBase):
    def read(self, infile):
        import sqlite3
        self._db = sqlite3.connect(infile.name)
        curs = self._db.cursor()

        curs.execute(
            'SELECT config_values.name, value FROM config_values WHERE site_name IS NULL')
        config = maybe_json_dict(curs)
        self.load_words_file(config.pop('words-file', None))
        self.set_defaults(self._get_site('default'))
        self.site_hashing.update(config.pop('site-hashing', {}))
        self.global_config = config

    def get_site_config(self, site):
        curs = self._db.cursor()
        curs.execute(
            'SELECT name, value FROM config_values WHERE site_name IS ?',
            (site,))
        return maybe_json_dict(curs)

    def _get_site(self, site):
        curs = self._db.cursor()
        curs.execute(
            'SELECT value FROM sites JOIN schemata USING (schema_id) WHERE site_name = ?',
            (site,))
        results = curs.fetchall()
        if not results:
            return None

        config = self.defaults.copy()
        config['schema'] = json.loads(results[0][0])
        config.update(self.get_site_config(site))

        self.fill_out_config(config)
        return config

    def add_site(self, name, schema_id):
        curs = self._db.cursor()
        curs.execute(
            'INSERT INTO sites (site_name, schema_id) VALUES (?, ?)',
            (name, schema_id))
        self._db.commit()

    def set_site_schema(self, name, schema_id):
        curs = self._db.cursor()
        curs.execute(
            'UPDATE sites SET schema_id = ? WHERE site_name = ?',
            (schema_id, name))
        self._db.commit()

    def remove_site(self, name):
        curs = self._db.cursor()
        curs.execute('DELETE FROM sites WHERE site_name = ?', (name,))
        curs.execute('DELETE FROM config_values WHERE site_name = ?', (name,))
        self._db.commit()

    def get_all_sites(self):
        curs = self._db.cursor()
        curs.execute(
            'SELECT site_name, name, value FROM config_values WHERE site_name IS NOT NULL')
        sites = collections.defaultdict(self.defaults.copy)
        for site, k, v in curs:
            sites[site][k] = maybe_json(v)

        curs.execute(
            'SELECT site_name, name, value FROM sites JOIN schemata USING (schema_id) WHERE site_name IS NOT NULL')
        for site, schema_name, schema in curs:
            sites[site]['schema-name'] = schema_name
            sites[site]['schema'] = json.loads(schema)

        for config in sites.values():
            self.fill_out_config(config)

        return dict(sites)

    def get_all_schemata(self):
        curs = self._db.cursor()
        curs.execute('SELECT name, value FROM schemata')
        return maybe_json_dict(curs)

    def get_schema(self, name):
        curs = self._db.cursor()
        curs.execute('SELECT schema_id, value FROM schemata WHERE name = ?', (name,))
        results = curs.fetchall()
        if not results:
            raise ValueError('there is no schema by the name %r' % (name,))
        return results[0]

    def add_schema(self, name, value):
        curs = self._db.cursor()
        curs.execute(
            'INSERT INTO schemata (name, value) VALUES (?, ?)',
            (name, jdumps(value)))
        self._db.commit()

    def remove_schema(self, schema_id):
        curs = self._db.cursor()
        curs.execute('SELECT site_name FROM sites WHERE schema_id = ?', (schema_id,))
        sites = sorted(site for site, in curs)
        if sites:
            raise ValueError(
                "can't delete this schema; at least one site is using it: %r" % (sites,))
        curs.execute('DELETE FROM schemata WHERE schema_id = ?', (schema_id,))
        self._db.commit()

    def set_schema_name(self, schema_id, newname):
        curs = self._db.cursor()
        curs.execute('UPDATE schemata SET name = ? WHERE schema_id = ?', (newname, schema_id))
        self._db.commit()

    def set_schema_value(self, schema_id, value):
        curs = self._db.cursor()
        curs.execute(
            'UPDATE schemata SET value = ? WHERE schema_id = ?',
            (jdumps(value), schema_id))
        self._db.commit()

    def get_config(self, site, name):
        curs = self._db.cursor()
        curs.execute(
            'SELECT value FROM config_values WHERE site_name IS ? AND name = ?',
            (site, name))
        results = curs.fetchall()
        if not results:
            raise ValueError('there is no config %r for %r' % (name, site))
        return maybe_json(results[0][0])

    def set_config(self, site, name, value):
        split_name = name.split('.')
        if len(split_name) > 1:
            try:
                base_value = self.get_config(site, split_name[0])
            except ValueError:
                base_value = {}
            nested_set(base_value, split_name[1:], value)
            name, new_value = split_name[0], base_value
        else:
            new_value = value
        curs = self._db.cursor()
        curs.execute(
            'DELETE FROM config_values WHERE site_name IS ? AND name = ?',
            (site, name,))
        if new_value is not None:
            curs.execute(
                'INSERT INTO config_values (site_name, name, value) VALUES (?, ?, ?)',
                (site, name, jdumps(new_value)))
        self._db.commit()


def load(infile):
    if infile.read(16) == b'SQLite format 3\x00':
        config = SqliteConfig()
    else:
        config = YAMLConfig()
    infile.seek(0)
    config.read(infile)
    return config



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import json
import sqlite3
import yaml


def main(yaml_file, sqlite_file):
    with open(yaml_file) as infile:
        loader = yaml.SafeLoader(infile)
        loader.check_node()
        loader.get_event()
        node = loader.compose_node(None, None)
        schema_names = dict(
            (json.dumps(list(map(loader.construct_sequence, v.value))), k)
            for k, v in loader.anchors.items())
        config = loader.construct_document(node)
    db = sqlite3.connect(sqlite_file)
    sites = config.pop('sites', {})
    config.pop('schemata', None)
    default_schema = sites['default']['schema']
    schemata = {}
    config_rows = [
        (None, k, json.dumps(v)) for k, v in config.items()]
    for site, site_config in sites.items():
        schema = json.dumps(site_config.pop('schema', default_schema))
        schemata.setdefault(schema, []).append(site)
        config_rows.extend(
            (site, k, json.dumps(v)) for k, v in site_config.items())
    for e, schema in enumerate(sorted(schemata)):
        schema_names.setdefault(schema, 'schema_%d' % e)
    curs = db.cursor()
    curs.executemany(
        'INSERT INTO schemata (name, value) VALUES (?, ?)',
        ((name, schema) for schema, name in schema_names.items()))
    curs.executemany(
        'INSERT INTO sites (site_name, schema_id) SELECT ?, schema_id FROM schemata WHERE value = ?',
        ((site, value)
         for value, sites in schemata.items()
         for site in sites))
    curs.executemany('INSERT INTO config_values (site_name, name, value) VALUES (?, ?, ?)', config_rows)
    db.commit()


if __name__ == '__main__':  # pragma: nocover
    import sys
    main(*sys.argv[1:])



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals, print_function

from passacre.config import load as load_config, SqliteConfig
from passacre.generator import hash_site
from passacre.schema import multibase_of_schema
from passacre.util import reify, dotify, nested_get, jdumps
from passacre import __version__, yaml2sqlite

import atexit
import collections
import json
from getpass import getpass
import math
import operator
import os
import sys
import time

try:
    import xerox
except ImportError:  # pragma: nocover
    xerox = None


if sys.version_info < (3,):  # pragma: nocover
    input = raw_input

if sys.version_info < (3, 3):  # pragma: nocover
    import passacre._argparse as argparse
else:  # pragma: nocover
    import argparse


def open_first(paths, mode='r', expanduser=None, open=open):
    if expanduser is None:
        expanduser = os.path.expanduser
    for path in paths:
        try:
            return open(expanduser(path), mode)
        except EnvironmentError:
            pass
    raise ValueError('no file in %r could be opened' % (paths,))

def prompt(query, input=input):
    sys.stderr.write(query)
    return input()

def prompt_password(confirm=False, getpass=getpass):
    password = getpass()
    if confirm and password != getpass('Confirm password: '):
        raise ValueError("passwords don't match")
    return password

def is_likely_hashed_site(site):
    return len(site) == 48 and '.' not in site

def site_sort_key(site):
    return not is_likely_hashed_site(site), site

def maybe_load_json(val):
    try:
        return json.loads(val)
    except ValueError:
        return val

schema_file = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'schema.sql')


class Passacre(object):
    atexit = atexit
    xerox = xerox
    prompt = staticmethod(prompt)
    _prompt_password = staticmethod(prompt_password)
    sleep = staticmethod(time.sleep)
    _load_config = staticmethod(load_config)

    _config = _config_file = None

    _subcommands = {
        'init': "initialize an sqlite config",
        'generate': "generate a password",
        'entropy': "display each site's password entropy",
        'site': ("actions on sites", {
            'add': "add a site to a config file",
            'remove': "remove a site from a config file",
            'set-schema': "change a site's schema",
            'hash': "hash a site's name",
        }),
        'schema': ("actions on schemata", {
            'add': "add a schema",
            'remove': "remove a schema",
            'set-value': "change a schema's value",
            'set-name': "change a schema's name",
        }),
        'config': "view/change global configuration",
    }

    @reify
    def config(self):
        if self._config is None:
            self.load_config(self._config_file)
        return self._config

    def load_config(self, config_fobj=None, expanduser=None):
        "Load the passacre configuration to ``self.config``."
        if config_fobj is None:
            config_fobj = open_first([
                '~/.config/passacre/passacre.sqlite',
                '~/.config/passacre/passacre.yaml',
                '~/.passacre.sqlite',
                '~/.passacre.yaml',
            ], 'rb', expanduser)
        with config_fobj:
            self._config = self._load_config(config_fobj)

    def prompt_password(self, confirm):
        if self.config.global_config.get('always-confirm-passwords'):
            confirm = True
        return self._prompt_password(confirm)


    def init_args(self, subparser):
        subparser.add_argument('path', nargs='?', default='~/.passacre.sqlite',
                               help='path of the config file to initialize (default: %(default)s)')
        subparser.add_argument('-y', '--from-yaml', type=argparse.FileType('rb'), metavar='YAML',
                               help='optional input YAML config file to convert from')

    def init_action(self, args):
        import sqlite3
        path = os.path.expanduser(args.path)
        db = sqlite3.connect(path)
        curs = db.cursor()
        with open(schema_file) as infile:
            schema = infile.read()
        curs.executescript(schema)

        if args.from_yaml:
            yaml2sqlite.main(args.from_yaml.name, path)
        else:
            config = SqliteConfig()
            config._db = db
            config.add_schema('32-printable', [[32, 'printable']])
            schema_id, _ = config.get_schema('32-printable')
            config.add_site('default', schema_id)


    def generate_args(self, subparser):
        subparser.add_argument('site', nargs='?',
                               help='site for which to generate a password')
        subparser.add_argument('-u', '--username',
                               help='username for the site')
        subparser.add_argument('-n', '--no-newline', action='store_true',
                               help="don't write a newline after the password")
        subparser.add_argument('-c', '--confirm', action='store_true',
                               help='confirm prompted password')
        if self.xerox is not None:
            subparser.add_argument('-C', '--copy', action='store_true',
                                   help='put the generated password on the clipboard')
            subparser.add_argument('-w', '--timeout', type=int, metavar='N',
                                   help='clear the clipboard after N seconds')

    def generate_action(self, args):
        "Generate a password."
        password = self.prompt_password(args.confirm)
        if args.site is None:
            args.site = self.prompt('Site: ')
        password = self.config.generate_for_site(
            args.username, password, args.site)
        if getattr(args, 'copy', False):  # since the argument might not exist
            sys.stderr.write('password copied.\n')
            self.xerox.copy(password)
            if args.timeout:
                self.atexit.register(self.xerox.copy, '')
                try:
                    self.sleep(args.timeout)
                except KeyboardInterrupt:
                    pass
        else:
            sys.stdout.write(password)
            if not args.no_newline:
                sys.stdout.write('\n')


    def entropy_args(self, subparser):
        subparser.add_argument('--schema', action='store_true',
                               help='show entropy by schema instead of by site')

    def entropy_action(self, args):
        """Display each site's potential password entropy in bits.

        This is the same thing as the base 2 logarithm of the number of
        possible passwords for a site.
        """

        entropy = [('schema' if args.schema else 'site', 'entropy', '(bits)'), ('', '', '')]
        if args.schema:
            pre_entropy = []
            for schema_name, schema in self.config.get_all_schemata().items():
                multibase = multibase_of_schema(schema, self.config.words)
                pre_entropy.append((schema_name, math.log(multibase.max_encodable_value + 1, 2)))
        else:
            default_site = self.config.get_site('default')
            pre_entropy = [
                (site, math.log(site_config['multibase'].max_encodable_value + 1, 2))
                for site, site_config in self.config.get_all_sites().items()
                if site_config['schema'] != default_site['schema'] or site == 'default'
            ]
        pre_entropy.sort(key=operator.itemgetter(1, 0), reverse=True)
        entropy.extend([site] + ('%0.2f' % bits).split('.')
                       for site, bits in pre_entropy)
        max_site_len, max_ibits_len, max_fbits_len = [
            len(max(x, key=len)) for x in zip(*entropy)]
        print()
        for e, (site, ibits, fbits) in enumerate(entropy):
            if e == 0:
                site = site.center(max_site_len)
            print('%*s   %*s%s%*s' % (
                -max_site_len, site, max_ibits_len, ibits,
                '.' if ibits.isdigit() else ' ', -max_fbits_len, fbits))


    def site_args(self, subparser):
        subparser.add_argument('--by-schema', action='store_true',
                               help='list sites organized by schema')

        hash_group = subparser.add_argument_group('for {add,remove,set-schema}')
        hash_group.add_argument('-a', '--hashed', action='store_true',
                                help='hash the site name')
        hash_group.add_argument('-c', '--confirm', action='store_true',
                                help='confirm prompted password')

    def site_action(self, args):
        "Perform an action on a site in a config file."

        sites = self.config.get_all_sites()
        if args.by_schema:
            sites_by_schema = collections.defaultdict(list)
            for site, site_config in sites.items():
                sites_by_schema[site_config['schema-name']].append(site)
            for schema in sorted(sites_by_schema):
                print('%s: %s' % (
                    schema, ', '.join(sorted(sites_by_schema[schema], key=site_sort_key))))
            return

        for site in sorted(sites, key=site_sort_key):
            site_config = sites[site]
            if 'schema-name' in site_config:
                print('%s: %s' % (site, site_config['schema-name']))
            else:
                print(site)


    def perhaps_hash_site(self, args):
        if args.hashed or (self.config.site_hashing['enabled'] == 'always' and args.site != 'default'):
            password = self.prompt_password(args.confirm)
            args.site = hash_site(password, args.site, self.config.site_hashing)


    def site_hash_args(self, subparser):
        subparser.add_argument('site', nargs='?',
                               help='the site to hash')
        subparser.add_argument('-m', '--method',
                               help='which hash method to use')
        subparser.add_argument('-n', '--no-newline', action='store_true',
                               help="don't write a newline after the hash")
        subparser.add_argument('-c', '--confirm', action='store_true',
                               help='confirm prompted password')

    def site_hash_action(self, args):
        """Hash a site.

        This is so that hostnames need not leak from a config file; the hashed
        site name is tried if the unhashed name doesn't exist.
        """

        password = self.prompt_password(args.confirm)
        if args.site is None:
            args.site = self.prompt('Site: ')
        config = self.config.site_hashing
        if args.method is not None:
            config['method'] = args.method
        sys.stdout.write(hash_site(password, args.site, config))
        if not args.no_newline:
            sys.stdout.write('\n')


    def site_add_args(self, subparser):
        subparser.add_argument('site', help='the name of the site')
        subparser.add_argument('schema', help='the schema to use')

    def site_add_action(self, args):
        schema_id, _ = self.config.get_schema(args.schema)
        self.perhaps_hash_site(args)
        self.config.add_site(args.site, schema_id)


    def site_remove_args(self, subparser):
        subparser.add_argument('site', help='the name of the site to remove')

    def site_remove_action(self, args):
        if args.site == 'default':
            sys.exit("can't remove the default site")
        self.perhaps_hash_site(args)
        self.config.remove_site(args.site)


    def site_set_schema_args(self, subparser):
        subparser.add_argument('site', help='the name of the site to update')
        subparser.add_argument('schema', help='the schema to use')

    def site_set_schema_action(self, args):
        schema_id, _ = self.config.get_schema(args.schema)
        self.perhaps_hash_site(args)
        self.config.set_site_schema(args.site, schema_id)


    def schema_action(self, args):
        "Perform an action on a schema in a config file."

        schemata = self.config.get_all_schemata()
        for schema in sorted(schemata):
            print('%s: %s' % (schema, jdumps(schemata[schema])))


    def schema_add_args(self, subparser):
        subparser.add_argument('name', help='the name of the schema')
        subparser.add_argument('value', help='the value of the schema')

    def schema_add_action(self, args):
        self.config.add_schema(args.name, json.loads(args.value))


    def schema_remove_args(self, subparser):
        subparser.add_argument('name', help='the name of the schema to remove')

    def schema_remove_action(self, args):
        schema_id, _ = self.config.get_schema(args.name)
        self.config.remove_schema(schema_id)


    def schema_set_name_args(self, subparser):
        subparser.add_argument('oldname', help='the schema to set the name of')
        subparser.add_argument('newname', help='the new name of the schema')

    def schema_set_name_action(self, args):
        schema_id, _ = self.config.get_schema(args.oldname)
        self.config.set_schema_name(schema_id, args.newname)


    def schema_set_value_args(self, subparser):
        subparser.add_argument('name', help='the name of the schema')
        subparser.add_argument('value', help='the new value for the schema')

    def schema_set_value_action(self, args):
        schema_id, _ = self.config.get_schema(args.name)
        self.config.set_schema_value(schema_id, json.loads(args.value))


    def config_args(self, subparser):
        subparser.add_argument('-s', '--site',
                               help='the site to operate on or omitted for global config')
        subparser.add_argument('-a', '--hashed', action='store_true',
                               help='hash the site name')
        subparser.add_argument('-c', '--confirm', action='store_true',
                               help='confirm prompted password')
        subparser.add_argument('name', nargs='?',
                               help='the config option to get or set or omitted to get all')
        subparser.add_argument('value', nargs='?',
                               help='the new value to set or omitted to get')

    def config_action(self, args):
        if args.site:
            self.perhaps_hash_site(args)
        if not args.name:
            for k, v in sorted(dotify(self.config.get_site_config(args.site))):
                print('%s: %s' % (k, jdumps(v)))
            return
        if not args.value:
            value = nested_get(
                self.config.get_site_config(args.site), args.name.split('.'))
            if value is not None:
                print(jdumps(value))
            return
        self.config.set_config(args.site, args.name, maybe_load_json(args.value))


    def build_subcommands(self, action_prefix, subparsers, subcommands):
        for subcommand, subcommand_help in subcommands.items():
            subsubcommand = False
            if isinstance(subcommand_help, tuple):
                subcommand_help, subsubcommands = subcommand_help
                subsubcommand = True
            subcommand_method = action_prefix + subcommand.replace('-', '_')
            action_method = getattr(self, '%s_action' % (subcommand_method,))
            subparser = subparsers.add_parser(
                subcommand, help=subcommand_help, description=action_method.__doc__)
            args_method = getattr(self, '%s_args' % (subcommand_method,), None)
            if args_method is not None:
                args_method(subparser)

            if subsubcommand:
                subaction_prefix = subcommand_method + '_'
                command = subaction_prefix + 'command'
                subsubparsers = subparser.add_subparsers(dest=command)
                self.build_subcommands(subaction_prefix, subsubparsers, subsubcommands)

    def build_parser(self):
        "Build an ``ArgumentParser`` from the defined subcommands."
        parser = argparse.ArgumentParser(prog='passacre')
        parser.add_argument('-V', '--version', action='version',
                            version='%(prog)s ' + __version__)
        parser.add_argument('-f', '--config', type=argparse.FileType('rb'),
                            help='specify a config file to use')
        subparsers = parser.add_subparsers(dest='command')
        self.build_subcommands('', subparsers, self._subcommands)
        return parser

    def find_action(self, args):
        if not args.command:
            return None
        command = ret = args.command.replace('-', '_')
        while True:
            next_command = command + '_command'
            subcommand = getattr(args, next_command, None)
            if subcommand is None:
                break
            command, ret = next_command, '%s_%s' % (ret, subcommand.replace('-', '_'))
        return ret

    def main(self, args=None):
        parser = self.build_parser()
        args = parser.parse_args(args)
        self._config_file = args.config
        action = self.find_action(args)
        if not action:
            parser.print_help()
            sys.exit(2)
        action_method = getattr(self, action + '_action')
        action_method(args)

def main(args=None):  # pragma: nocover
    Passacre().main(args)



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals

import math
import string
import sys

from passacre.multibase import MultiBase

_some_nulls = b'\x00' * 1024
_site_multibase = MultiBase([string.ascii_letters + string.digits + '-_'] * 48)
_site_multibase_bits = 288


if sys.version_info > (3,):  # pragma: nocover
    def perhaps_encode(s):
        return s.encode()
else:  # pragma: nocover
    perhaps_encode = lambda x: x


def generate(username, password, site, options):
    """Generate a password with the passacre method.

    1. A string is generated from ``username:`` (if a username is specified),
       contacenated with ``password:site``, concatentated with 1024 null bytes
       for every iteration.
    2. A pseudo-random number generator is initialized using the string as a
       seed.
    3. The PRNG is asked for an integer below the maximum value that
       ``multibase`` can encode.
    4. That integer is encoded with
       ``multibase`` and the encoded value is returned.
    """

    multibase = options['multibase']
    prng = build_prng(username, password, site, options)

    required_bytes = int(math.ceil(
        math.log(multibase.max_encodable_value + 1, 256)))
    required_bits = required_bytes * 8
    while True:
        password_value = prng.getrandbits(required_bits)
        if password_value <= multibase.max_encodable_value:
            break
    return multibase.encode(password_value)


# XXX: get this included in pyskein
def _patch_skein_random(prng):
    def getrandbits(n_bits):
        "Patched in implementation of getrandbits for skein."
        if n_bits <= 0:
            raise ValueError('number of bits must be greater than zero')
        n_bytes = (n_bits + 7) // 8
        val = int.from_bytes(prng.read(n_bytes), 'little')
        return val >> (n_bytes * 8 - n_bits)
    prng.getrandbits = getrandbits
    return prng


# XXX: refactor into a class per method?
def build_prng(username, password, site, options):
    method = options['method']
    iterations = options['iterations']
    seed = (
        (perhaps_encode(username) + b':' if username else b'')
        + perhaps_encode(password) + b':'
        + site.encode('idna')
        + (_some_nulls * iterations))

    if method == 'keccak':
        import keccak
        sponge = keccak.Sponge(64, 1536)
        sponge.absorb(seed)
        return keccak.SpongeRandom(sponge)

    elif method == 'skein':
        import skein
        return _patch_skein_random(skein.Random(seed))

    else:
        raise ValueError('invalid method: %r' % (method,))

def hash_site(password, site, options):
    prng = build_prng(None, password, site, options)
    return _site_multibase.encode(prng.getrandbits(_site_multibase_bits))



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals

import sys

import pytest

from passacre import schema


if sys.version_info < (3,):
    string_prefix, string_padding = 'u', ' '
else:
    string_prefix = string_padding = ''


def test_schema_parse_error_single_item_list():
    with pytest.raises(schema.ParseError) as excinfo:
        schema.parse_items([['word']])
    assert str(excinfo.value) == """
whilst parsing the items: [[{0}'word']]
whilst parsing an item:    [{0}'word']
expected an array with at least two elements; got [{0}'word']""".format(string_prefix)

def test_schema_parse_error_wrong_type_in_items():
    with pytest.raises(schema.ParseError) as excinfo:
        schema.parse_items(None)
    assert str(excinfo.value) == """
whilst parsing the items: None
whilst parsing the value: None
expected an array; got None"""

def test_schema_parse_error_wrong_type_in_item():
    with pytest.raises(schema.ParseError) as excinfo:
        schema.parse_items([None])
    assert str(excinfo.value) == """
whilst parsing the items:       [None]
whilst parsing an item:          None
whilst parsing character sets:   None
whilst parsing a character set:  None
whilst parsing the value:        None
expected a string; got None"""

def test_schema_parse_error_wrong_type_in_delimiter():
    with pytest.raises(schema.ParseError) as excinfo:
        schema.parse_items([[None, 1]])
    assert str(excinfo.value) == """
whilst parsing the items:               [[None, 1]]
whilst parsing an item:                  [None, 1]
whilst parsing a count and items array:  [None, 1]
whilst parsing the value:                 None
expected a string; got None"""

def test_schema_parse_error_in_non_zeroth_item():
    with pytest.raises(schema.ParseError) as excinfo:
        schema.parse_items([[' ', None]])
    assert str(excinfo.value) == """
whilst parsing the items:       [[{0}' ', None]]
whilst parsing an item:          [{0}' ', None]
whilst parsing character sets:   [{0}' ', None]
whilst parsing a character set:   {1}     None
whilst parsing the value:         {1}     None
expected a string; got None""".format(string_prefix, string_padding)



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals

import pytest
import py.path
import unittest

from passacre import application
from passacre.test.util import excinfo_arg_0


datadir = py.path.local(__file__).dirpath('data')


def fake_open(path, mode):
    if 'fail' in path:
        raise IOError()
    return path, mode

def fake_expanduser(path):
    return path.replace('~', '.')

def read_out(capsys, app, *args):
    app.main(args)
    out, err = capsys.readouterr()
    assert not err
    return out

def read_err(capsys, app, *args):
    app.main(args)
    out, err = capsys.readouterr()
    assert not out
    return err


def test_open_first():
    assert application.open_first([
        'fail-spam',
        'fail-spam',
        'eggs',
        'spam',
    ], open=fake_open) == ('eggs', 'r')
    assert application.open_first([
        '~/fail-spam',
        '~/fail-spam',
        '~/eggs',
        '~/spam',
    ], open=fake_open, expanduser=fake_expanduser) == ('./eggs', 'r')
    assert application.open_first([
        'fail-spam',
        'fail-spam',
        'eggs',
        'spam',
    ], 'w', open=fake_open) == ('eggs', 'w')
    with pytest.raises(ValueError):
        application.open_first(['fail-spam', 'fail-eggs'], open=fake_open)
    with pytest.raises(ValueError):
        application.open_first([
            '~/fail-spam',
            '~/fail-eggs'
        ], open=fake_open, expanduser=fake_expanduser)
    with pytest.raises(ValueError):
        application.open_first([], open=fake_open)


def test_prompt(capsys):
    assert application.prompt('foo: ', input=lambda: 'bar') == 'bar'
    out, err = capsys.readouterr()
    assert not out
    assert err == 'foo: '


def fake_getpass(outputs):
    def getpass(prompt=None):
        assert prompt in outputs
        return outputs[prompt]
    return getpass

def test_prompt_password():
    assert 'bar' == application.prompt_password(False, getpass=fake_getpass({None: 'bar'}))
    assert 'bar' == application.prompt_password(
        True, getpass=fake_getpass({None: 'bar', 'Confirm password: ': 'bar'}))
    with pytest.raises(ValueError):
        application.prompt_password(
            True, getpass=fake_getpass({None: 'foo', 'Confirm password: ': 'bar'}))


class FakeXerox(object):
    def __init__(self):
        self.copied = []

    def copy(self, arg):
        self.copied.append(arg)

class FakeSleep(object):
    def __init__(self, interrupt=False):
        self.duration = None
        self.interrupt = interrupt

    def __call__(self, duration):
        self.duration = duration
        if self.interrupt:
            raise KeyboardInterrupt()

class FakeAtexit(object):
    def __init__(self):
        self.calls = []

    def register(self, f, *a, **kw):
        self.calls.append((f, a, kw))

    def call_all(self):
        for f, a, kw in self.calls:
            f(*a, **kw)


class ApplicationTestCaseMixin(object):
    config_dir = None
    hashed_site = None
    hashed_password = None

    @pytest.fixture(autouse=True)
    def init_capsys(self, capsys):
        self.capsys = capsys

    def setUp(self):
        datadir.join(self.config_dir).chdir()
        self.app = application.Passacre()
        _load_config = self.app.load_config
        self.app.load_config = lambda f=None: _load_config(f, expanduser=fake_expanduser)
        self.confirmed_password = False
        self.app._prompt_password = self.prompt_password
        self.password = 'passacre'

    def prompt_password(self, confirm):
        self.confirmed_password = confirm
        return self.password


    def test_generate_with_newline(self):
        out = read_out(self.capsys, self.app, 'generate', 'example.com')
        assert not self.confirmed_password
        assert out == self.hashed_password + '\n'

    def test_generate_no_newline(self):
        out = read_out(self.capsys, self.app, 'generate', '-n', 'example.com')
        assert out == self.hashed_password

    def test_generate_prompt_site(self):
        self.app.prompt = lambda _: 'example.com'
        out = read_out(self.capsys, self.app, 'generate')
        assert out == self.hashed_password + '\n'

    def test_generate_confirm_password(self):
        out = read_out(self.capsys, self.app, 'generate', 'example.com', '-c')
        assert self.confirmed_password
        assert out == self.hashed_password + '\n'

    def test_generate_copying(self):
        self.app.xerox = FakeXerox()
        err = read_err(self.capsys, self.app, 'generate', 'example.com', '-C')
        assert err == 'password copied.\n'
        assert self.app.xerox.copied == [self.hashed_password]

    def test_generate_copy_timeout(self):
        self.app.xerox = FakeXerox()
        self.app.sleep = FakeSleep()
        self.app.atexit = FakeAtexit()
        err = read_err(self.capsys, self.app, 'generate', 'example.com', '-C', '-w', '10')
        assert err == 'password copied.\n'
        assert self.app.xerox.copied == [self.hashed_password]
        assert self.app.sleep.duration == 10
        self.app.atexit.call_all()
        assert self.app.xerox.copied == [self.hashed_password, '']

    def test_generate_copy_timeout_interrupted(self):
        self.app.xerox = FakeXerox()
        self.app.sleep = FakeSleep(interrupt=True)
        self.app.atexit = FakeAtexit()
        err = read_err(self.capsys, self.app, 'generate', 'example.com', '-C', '-w', '10')
        assert err == 'password copied.\n'
        assert self.app.xerox.copied == [self.hashed_password]
        assert self.app.sleep.duration == 10
        self.app.atexit.call_all()
        assert self.app.xerox.copied == [self.hashed_password, '']


    def test_entropy(self):
        out = read_out(self.capsys, self.app, 'entropy')
        # stripping trailing whitespace is a bitch
        out = '\n'.join(line.rstrip() for line in out.splitlines())
        assert out == """
                      site                         entropy (bits)

default                                                209.75
becu.org                                               190.53
fidelity.com                                           101.72
schwab.com                                              47.63
still.further.example.com                               39.68
further.example.com                                     26.58
example.com                                             26.58
%s        13.29""" % (self.hashed_site,)


    def test_site_hash_with_newline(self):
        out = read_out(self.capsys, self.app, 'site', 'hash', 'hashed.example.com')
        assert not self.confirmed_password
        assert out == self.hashed_site + '\n'

    def test_site_hash_no_newline(self):
        out = read_out(self.capsys, self.app, 'site', 'hash', '-n', 'hashed.example.com')
        assert out == self.hashed_site

    def test_site_hash_prompt_site(self):
        self.app.prompt = lambda _: 'hashed.example.com'
        out = read_out(self.capsys, self.app, 'site', 'hash')
        assert out == self.hashed_site + '\n'

    def test_site_hash_confirm_password(self):
        out = read_out(self.capsys, self.app, 'site', 'hash', 'hashed.example.com', '-c')
        assert self.confirmed_password
        assert out == self.hashed_site + '\n'

    def test_site_hash_overridden_method_keccak(self):
        out = read_out(self.capsys, self.app, 'site', 'hash', 'hashed.example.com', '-m', 'keccak')
        assert out == 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4\n'

    @pytest.mark.skipif("sys.version_info < (3,)")
    def test_site_hash_overridden_method_skein(self):
        out = read_out(self.capsys, self.app, 'site', 'hash', 'hashed.example.com', '-m', 'skein')
        assert out == 'UYfDoAN9nYMdxCYtgKenzjhbc9eonu3w92ec3SAA5UbT1J3L\n'


    def test_no_args(self):
        with pytest.raises(SystemExit) as excinfo:
            self.app.main([])
        assert excinfo_arg_0(excinfo) == 2


class SqliteTestCaseMixin(object):
    def test_schema_entropy(self):
        out = read_out(self.capsys, self.app, 'entropy', '--schema')
        out = '\n'.join(line.rstrip() for line in out.splitlines())
        assert out == """
 schema    entropy (bits)

schema_5       209.75
schema_3       190.53
schema_6       101.72
schema_1        47.63
schema_4        39.68
schema_2        26.58
schema_0        26.58
schema_7        13.29"""


class KeccakTestCaseMixin(ApplicationTestCaseMixin):
    hashed_site = 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4'
    hashed_password = 'Abaris abaisance abashedness abarticulation'

class KeccakYAMLTestCase(KeccakTestCaseMixin, unittest.TestCase):
    config_dir = 'keccak-yaml'

class KeccakSqliteTestCase(KeccakTestCaseMixin, SqliteTestCaseMixin, unittest.TestCase):
    config_dir = 'keccak-sqlite'


@pytest.mark.skipif("sys.version_info < (3,)")
class SkeinTestCaseMixin(ApplicationTestCaseMixin):
    hashed_site = 'UYfDoAN9nYMdxCYtgKenzjhbc9eonu3w92ec3SAA5UbT1J3L'
    hashed_password = 'abactor abattoir abashedly abaca'

class SkeinYAMLTestCase(SkeinTestCaseMixin, unittest.TestCase):
    config_dir = 'skein-yaml'

class SkeinSqliteTestCase(SkeinTestCaseMixin, SqliteTestCaseMixin, unittest.TestCase):
    config_dir = 'skein-sqlite'


@pytest.fixture
def app():
    app = application.Passacre()
    app.load_config(datadir.join('keccak.sqlite').open('rb'))
    return app

@pytest.fixture
def init_app(app, tmpdir):
    dbpath = tmpdir.join('test.db').strpath
    app.main(['init', dbpath])
    return dbpath, application.Passacre()

def test_init_config(init_app, capsys):
    dbpath, app = init_app
    out = read_out(capsys, app, '-f', dbpath, 'config')
    assert out == ''

def test_init_schemata(init_app, capsys):
    dbpath, app = init_app
    out = read_out(capsys, app, '-f', dbpath, 'schema')
    assert out == '32-printable: [[32, "printable"]]\n'

def test_init_sites(init_app, capsys):
    dbpath, app = init_app
    out = read_out(capsys, app, '-f', dbpath, 'site')
    assert out == 'default: 32-printable\n'


@pytest.fixture
def yaml2sqlite_app(tmpdir):
    tmp_app = application.Passacre()
    dbpath = tmpdir.join('test.db')
    tmp_app.main(
        ['init', '-y', datadir.join('keccak.yaml').strpath, dbpath.strpath])
    app = application.Passacre()
    app.load_config(dbpath.open('rb'))
    return app

def test_init_from_yaml(yaml2sqlite_app, capsys):
    app = yaml2sqlite_app
    assert read_out(capsys, app, 'schema') == """12-alphanumeric: [[12, "alphanumeric"]]
32-alphanumeric: [[32, "alphanumeric"]]
32-printable: [[32, "printable"]]
8-alphanumeric: [[8, "alphanumeric"]]
schema_0: ["printable", [", ", 4, "word"], "printable"]
schema_1: [[" ", 2, "word"]]
schema_2: [[" ", 4, "word"]]
schema_3: [[", ", 4, "word"]]
schema_4: [[16, ["alphanumeric", "\\"%'()+,-/:;<=>?\\\\ ^_|"]]]
"""

def test_init_from_yaml_sites(yaml2sqlite_app, capsys):
    app = yaml2sqlite_app
    assert read_out(capsys, app, 'site') == """gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4: schema_1
becu.org: 32-alphanumeric
default: 32-printable
example.com: schema_2
fhcrc.org: 32-printable
fidelity.com: schema_4
further.example.com: schema_3
schwab.com: 8-alphanumeric
still.further.example.com: schema_0
"""

def test_init_from_yaml_config(yaml2sqlite_app, capsys):
    app = yaml2sqlite_app
    assert read_out(capsys, app, 'config') == """site-hashing.enabled: true
site-hashing.iterations: 10
words-file: "words"
"""

def test_init_from_yaml_site_config(yaml2sqlite_app, capsys):
    app = yaml2sqlite_app
    assert read_out(capsys, app, 'config', '-s', 'default') == """iterations: 10
method: "keccak"
"""


def test_site_yaml(app, capsys):
    app.load_config(datadir.join('keccak.yaml').open('rb'))
    out = read_out(capsys, app, 'site')
    assert out == """gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4
becu.org
default
example.com
fhcrc.org
fidelity.com
further.example.com
schwab.com
still.further.example.com
"""

def test_site_sqlite(app, capsys):
    out = read_out(capsys, app, 'site')
    assert out == """gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4: schema_7
becu.org: schema_3
default: schema_5
example.com: schema_0
fhcrc.org: schema_5
fidelity.com: schema_6
further.example.com: schema_2
schwab.com: schema_1
still.further.example.com: schema_4
"""

def test_site_sqlite_by_schema(app, capsys):
    out = read_out(capsys, app, 'site', '--by-schema')
    assert out == """schema_0: example.com
schema_1: schwab.com
schema_2: further.example.com
schema_3: becu.org
schema_4: still.further.example.com
schema_5: default, fhcrc.org
schema_6: fidelity.com
schema_7: gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4
"""


def copy_app(name, app, tmpdir):
    datadir.join(name).copy(tmpdir)
    db_copy = tmpdir.join(name)
    app.load_config(db_copy.open('rb'))
    app._prompt_password = lambda confirm: 'passacre'
    return app

@pytest.fixture
def mutable_app(app, tmpdir):
    return copy_app('keccak.sqlite', app, tmpdir)

def test_site_add(mutable_app, capsys):
    app = mutable_app
    assert '\nexample.org: schema_0\n' not in read_out(capsys, app, 'site')
    app.main(['site', 'add', 'example.org', 'schema_0'])
    assert '\nexample.org: schema_0\n' in read_out(capsys, app, 'site')

def test_site_add_hashed(mutable_app, capsys):
    app = mutable_app
    assert 'KE76ybZ-sO7o4iS944E_mo_jTQPCjzifFjyRELZS-RuDbcGu: schema_0\n' not in read_out(
        capsys, app, 'site')
    app.main(['site', '-a', 'add', 'example.org', 'schema_0'])
    assert 'KE76ybZ-sO7o4iS944E_mo_jTQPCjzifFjyRELZS-RuDbcGu: schema_0\n' in read_out(
        capsys, app, 'site')

def test_site_set_schema(mutable_app, capsys):
    app = mutable_app
    assert '\nexample.com: schema_0\n' in read_out(capsys, app, 'site')
    app.main(['site', 'set-schema', 'example.com', 'schema_7'])
    assert '\nexample.com: schema_7\n' in read_out(capsys, app, 'site')

def test_site_set_schema_hashed(mutable_app, capsys):
    app = mutable_app
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4: schema_7\n' in read_out(capsys, app, 'site')
    app.main(['site', '-a', 'set-schema', 'hashed.example.com', 'schema_0'])
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4: schema_0\n' in read_out(capsys, app, 'site')

def test_site_remove(mutable_app, capsys):
    app = mutable_app
    assert '\nexample.com:' in read_out(capsys, app, 'site')
    app.main(['site', 'remove', 'example.com'])
    assert '\nexample.com:' not in read_out(capsys, app, 'site')

def test_site_remove_hashed(mutable_app, capsys):
    app = mutable_app
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4:' in read_out(capsys, app, 'site')
    app.main(['site', '-a', 'remove', 'hashed.example.com'])
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4:' not in read_out(capsys, app, 'site')

def test_site_remove_default_fails(app):
    with pytest.raises(SystemExit) as excinfo:
        app.main(['site', 'remove', 'default'])
    assert excinfo_arg_0(excinfo)

def test_config(app, capsys):
    out = read_out(capsys, app, 'config')
    assert out == """site-hashing.enabled: true
site-hashing.iterations: 10
words-file: "words"
"""

    out = read_out(capsys, app, 'config', 'site-hashing.enabled')
    assert out == 'true\n'

    out = read_out(capsys, app, 'config', 'site-hashing')
    assert out == '{"enabled": true, "iterations": 10}\n'

    out = read_out(capsys, app, 'config', '-s', 'fhcrc.org')
    assert out == 'increment: 5\n'

    out = read_out(capsys, app, 'config', '-s', 'fhcrc.org', 'increment')
    assert out == '5\n'

def get_config_name(app, capsys, site, name):
    args = ['config']
    if site is not None:
        args.extend(['-s', site])
    args.append(name)
    app.main(args)
    out, err = capsys.readouterr()
    assert not err
    return out.rstrip('\n')

def test_config_set(mutable_app, capsys):
    app = mutable_app

    app.main(['config', 'words-file', 'foo'])
    assert get_config_name(app, capsys, None, 'words-file') == '"foo"'

    app.main(['config', 'words-file', 'null'])
    assert get_config_name(app, capsys, None, 'words-file') == ''

    app.main(['config', 'spam.spam.spam.eggs', 'spam'])
    assert get_config_name(app, capsys, None, 'spam') == '{"spam": {"spam": {"eggs": "spam"}}}'

    app.main(['config', 'spam.spam.spam.spam', 'spam'])
    assert get_config_name(app, capsys, None, 'spam') == (
        '{"spam": {"spam": {"eggs": "spam", "spam": "spam"}}}')

    app.main(['config', '-s', 'fhcrc.org', 'increment', '6'])
    assert get_config_name(app, capsys, 'fhcrc.org', 'increment') == '6'

    app.main(['config', '-s', 'example.com', 'increment', '7'])
    assert get_config_name(app, capsys, 'example.com', 'increment') == '7'

    app.main(['config', '-as', 'hashed.example.com', 'increment', '8'])
    assert get_config_name(
        app, capsys, 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4', 'increment') == '8'

    app.main(['config', '-cas', 'hashed.example.com', 'increment', '9'])
    assert get_config_name(
        app, capsys, 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4', 'increment') == '9'

def test_schema(app, capsys):
    app.main(['schema'])
    out, err = capsys.readouterr()
    assert not err
    assert out == r"""schema_0: [[" ", 4, "word"]]
schema_1: [[8, "alphanumeric"]]
schema_2: [[", ", 4, "word"]]
schema_3: [[32, "alphanumeric"]]
schema_4: ["printable", [", ", 4, "word"], "printable"]
schema_5: [[32, "printable"]]
schema_6: [[16, ["alphanumeric", "\"%'()+,-/:;<=>?\\ ^_|"]]]
schema_7: [[" ", 2, "word"]]
"""

def test_schema_add(mutable_app, capsys):
    app = mutable_app
    app.main(['schema', 'add', 'schema_10', '["alphanumeric", [10, "digit"]]'])
    assert '\nschema_10: ["alphanumeric", [10, "digit"]]\n' in read_out(capsys, app, 'schema')

def test_schema_set_value(mutable_app, capsys):
    app = mutable_app
    app.main(['schema', 'set-value', 'schema_0', '[[10, "digit"]]'])
    assert 'schema_0: [[10, "digit"]]\n' in read_out(capsys, app, 'schema')

def test_schema_set_name(mutable_app, capsys):
    app = mutable_app
    assert 'schema_10:' not in read_out(capsys, app, 'schema')
    app.main(['schema', 'set-name', 'schema_0', 'schema_10'])
    assert 'schema_10:' in read_out(capsys, app, 'schema')

def test_schema_remove(mutable_app, capsys):
    app = mutable_app
    app.main(['schema', 'add', 'schema_10', '["alphanumeric", [10, "digit"]]'])
    assert 'schema_10:' in read_out(capsys, app, 'schema')
    app.main(['schema', 'remove', 'schema_10'])
    assert 'schema_10:' not in read_out(capsys, app, 'schema')


@pytest.fixture
def always_hash_app(app, tmpdir):
    return copy_app('always-hash.sqlite', app, tmpdir)

def test_always_hash_site_add(always_hash_app, capsys):
    app = always_hash_app
    assert 'KE76ybZ-sO7o4iS944E_mo_jTQPCjzifFjyRELZS-RuDbcGu: schema_0\n' not in read_out(
        capsys, app, 'site')
    app.main(['site', 'add', 'example.org', 'schema_0'])
    assert 'KE76ybZ-sO7o4iS944E_mo_jTQPCjzifFjyRELZS-RuDbcGu: schema_0\n' in read_out(
        capsys, app, 'site')

def test_always_hash_site_remove(always_hash_app, capsys):
    app = always_hash_app
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4: schema_7\n' in read_out(
        capsys, app, 'site')
    app.main(['site', 'remove', 'hashed.example.com'])
    assert 'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4' not in read_out(
        capsys, app, 'site')

def test_always_hash_site_config(always_hash_app, capsys):
    app = always_hash_app
    assert read_out(capsys, app, 'config', '-s', 'hashed.example.com') == 'foo: "bar"\n'
    app.main(['config', '-s', 'hashed.example.com', 'foo', 'spam'])
    assert read_out(capsys, app, 'config', '-s', 'hashed.example.com') == 'foo: "spam"\n'


@pytest.fixture
def always_confirm_app(app, tmpdir):
    app = copy_app('always-confirm.sqlite', app, tmpdir)
    def prompt_password(confirm):
        app._confirmed_password = confirm
        return 'passacre'
    app._prompt_password = prompt_password
    return app

def test_always_confirm_site_add(always_confirm_app):
    app = always_confirm_app
    app.main(['site', '-a', 'add', 'example.org', 'schema_0'])
    assert app._confirmed_password

def test_always_confirm_site_remove(always_confirm_app):
    app = always_confirm_app
    app.main(['site', '-a', 'remove', 'hashed.example.com'])
    assert app._confirmed_password

def test_always_confirm_site_config(always_confirm_app):
    app = always_confirm_app
    app.main(['config', '-a', '-s', 'hashed.example.com'])
    assert app._confirmed_password


@pytest.fixture
def nonextant_words_app(app, tmpdir):
    return copy_app('nonextant-words.sqlite', app, tmpdir)

def test_nonextant_words_warns(nonextant_words_app):
    app = nonextant_words_app
    with pytest.raises(ValueError):
        app.main(['generate', 'example.com'])



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import string

from passacre.multibase import MultiBase


_word = object()
character_classes = {
    'printable': string.digits + string.ascii_letters + string.punctuation,
    'alphanumeric': string.digits + string.ascii_letters,
    'digit': string.digits,
    'letter': string.ascii_letters,
    'lowercase': string.ascii_lowercase,
    'uppercase': string.ascii_uppercase,
}


try:
    unicode
except NameError:  # pragma: nocover
    unicode = str

string_types = unicode, str


class ParseError(Exception):
    def __init__(self, got, expected):
        super(ParseError, self).__init__(got, expected)
        self.got = got
        self.expected = expected
        self.parse_stack = []

    def __str__(self):
        whats, xs, indices = zip(*reversed(self.parse_stack))
        whats_max_len = max(len(what) for what in whats) + 1
        whats = [(what + ':').ljust(whats_max_len) for what in whats]
        spaced_xs = []
        spaces = 0
        prev = None
        for x, index in zip(xs, indices):
            if index is not None:
                spaces += 1
                for e in range(index):
                    spaces += 2 + len(repr(prev[e]))
            spaced_xs.append(' ' * spaces + repr(x))
            prev = x
        repaired = zip(whats, spaced_xs)

        return '\n%s\nexpected %s; got %r' % (
            '\n'.join('whilst parsing %s %s' % x for x in repaired),
            self.expected, self.got)


def trace_parse(what):
    def deco(f):
        def wrap(x, *a, **kw):
            index = kw.pop('_index', None)
            try:
                return f(x, *a)
            except ParseError as e:
                e.parse_stack.append((what, x, index))
                raise
        return wrap
    return deco

@trace_parse('the value')
def parse_type(x, type_, expected):
    if not isinstance(x, type_):
        raise ParseError(x, expected)
    return x

@trace_parse('a character set')
def parse_character_set(x):
    x = parse_type(x, string_types, 'a string')
    return character_classes.get(x, x)

@trace_parse('character sets')
def parse_character_sets(x):
    if x == 'word':
        return [_word]
    elif isinstance(x, list):
        return [''.join(parse_character_set(y, _index=e) for e, y in enumerate(x))]
    else:
        return [parse_character_set(x)]

@trace_parse('a count and items array')
def parse_counted_item(x):
    delimiter = ''
    if isinstance(x[0], int):
        count = parse_type(x[0], int, 'a number', _index=0)
        start = 1
    else:
        delimiter = parse_type(x[0], string_types, 'a string', _index=0)
        count = parse_type(x[1], int, 'a number', _index=1)
        start = 2
    each_item = [z
                 for e, y in enumerate(x[start:], start=start)
                 for z in parse_item(y, _index=e)]
    items = []
    for e in range(count):
        if e != 0 and delimiter:
            items.append([delimiter])
        items.extend(each_item)
    return items

@trace_parse('an item')
def parse_item(x):
    if isinstance(x, list):
        if len(x) < 2:
            raise ParseError(x, 'an array with at least two elements')
        if isinstance(x[0], int) or isinstance(x[1], int):
            return parse_counted_item(x)
    return parse_character_sets(x)

@trace_parse('the items')
def parse_items(x):
    items = [parse_item(y, _index=e) for e, y in enumerate(parse_type(x, list, 'an array'))]
    return [y for subitems in items for y in subitems]


def multibase_of_schema(schema, words):
    "Convert a password schema from decoded YAML to a ``MultiBase``."
    items = parse_items(schema)
    if words is None and _word in items:
        raise ValueError("can't use a schema with 'word' without a words file")
    items = [words if item is _word else item for item in items]
    return MultiBase(items)



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import sys


if sys.version_info < (2, 7):
    def excinfo_arg_0(excinfo):
        return excinfo.value
else:
    def excinfo_arg_0(excinfo):
        return excinfo.value.args[0]



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from __future__ import unicode_literals

import os
import pytest
import unittest

from passacre import config


datadir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')

class ConfigTestCaseMixin(object):
    password = 'passacre'
    config_file = None
    method = None

    expected_passwords = {}
    expected_username_passwords = {}

    expected_sites = {
        'default': {
            'iterations': 10, 'schema': [[32, 'printable']],
        },
        'becu.org': {
            'iterations': 10, 'schema': [[32, 'alphanumeric']],
        },
        'example.com': {
            'iterations': 10, 'schema': [[' ', 4, 'word']],
        },
        'fhcrc.org': {
            'increment': 5, 'iterations': 15, 'schema': [[32, 'printable']],
        },
        'fidelity.com': {
            'iterations': 10,
            'schema': [[16, ['alphanumeric', '"%\'()+,-/:;<=>?\\ ^_|']]],
        },
        'further.example.com': {
            'iterations': 10, 'schema': [[', ', 4, 'word']],
        },
        'schwab.com': {
            'iterations': 10, 'schema': [[8, 'alphanumeric']],
        },
        'still.further.example.com': {
            'iterations': 10,
            'schema': ['printable', [', ', 4, 'word'], 'printable'],
        }
    }
    extra_expected_sites = {}

    maxDiff = None

    def setUp(self):
        os.chdir(datadir)
        self.config = config.load(open(self.config_file, 'rb'))

    def test_expected_passwords(self):
        for site, expected in self.expected_passwords.items():
            self.assertEqual(
                expected, self.config.generate_for_site(None, self.password, site))

    def test_expected_username_passwords(self):
        for (username, site), expected in self.expected_username_passwords.items():
            self.assertEqual(
                expected, self.config.generate_for_site(username, self.password, site))

    def test_get_all_sites(self):
        sites = self.config.get_all_sites()
        for site, config in list(sites.items()):
            config.pop('multibase', None)
            config.pop('schema-name', None)
            self.assertEqual(self.method, config.pop('method'))
            if site in self.extra_expected_sites:
                self.assertEqual(
                    self.extra_expected_sites[site], sites.pop(site))
        self.assertEqual(self.expected_sites, sites)


class KeccakTestCaseMixin(ConfigTestCaseMixin):
    method = 'keccak'
    expected_passwords = {
        'becu.org': 'qRnda94q1srpHWCjaQUTDobnxIgJkKlO',
        'fhcrc.org': 's"E;JE*?&md?19{2}<4Q.Rf):X?BUe^0',
        'schwab.com': 'jRWs2Wzl',
        'fidelity.com': 'B2M4=KQ2yzYLacW1',
        'example.com': 'Abaris abaisance abashedness abarticulation',
        'further.example.com': 'a, abaton, abatement, abask',
        'still.further.example.com': ')abased, abaculus, Abaris, abature=',
        'hashed.example.com': 'Abasgi abatement',
        'default.example.com': '-1`mCLq}HF?t#unm>dJC8k)FkLO-jcy8',
    }
    expected_username_passwords = {
        ('passacre', 'schwab.com'): 'JkbefmM3',
        ('passacre', 'fhcrc.org'): 'r(O.vF`Avz44Z;$ZvQ..4LqN/CEz/*#?',
        ('passacre', 'hashed.example.com'): 'Aaronic Abassin',
        ('passacre', 'default.example.com'): '''RN*6~h'Jc-A"Ro-OefetOVes~tR<~=[K''',
    }
    extra_expected_sites = {
        'gN7y2jQ72IbdvQZxrZLNmC4hrlDmB-KZnGJiGpoB4VEcOCn4': {
            'iterations': 10, 'schema': [[' ', 2, 'word']]
        },
    }


class KeccakYAMLTestCase(KeccakTestCaseMixin, unittest.TestCase):
    config_file = 'keccak.yaml'

class KeccakSqliteTestCase(KeccakTestCaseMixin, unittest.TestCase):
    config_file = 'keccak.sqlite'


@pytest.mark.skipif("sys.version_info < (3,)")
class SkeinTestCaseMixin(ConfigTestCaseMixin):
    method = 'skein'
    expected_passwords = {
        'becu.org': 'GEifpmhAzCtbTHBdj40B215CTXZATzzy',
        'fhcrc.org': 'C^#l04.LnhZ-m4}]6&RPMWYS#oW2_E)U',
        'schwab.com': '2CFNqPya',
        'fidelity.com': '9DT"bYHp4Gch\\"jL',
        'example.com': 'abactor abattoir abashedly abaca',
        'further.example.com': 'abandonable, aal, abalienate, abampere',
        'still.further.example.com': ';abate, abaciscus, abandonable, abalonea',
        'hashed.example.com': 'aardvark Abadite',
        'default.example.com': '~N8q0u$fYwnpD{VITu^F*,-NO1~PV>m&',
    }
    expected_username_passwords = {
        ('passacre', 'schwab.com'): 'sKvqo9Y0',
        ('not-passacre', 'schwab.com'): 'JlOdJ7KC',
        ('passacre', 'fhcrc.org'): '''~,HDW+W9@hg{'pU*P"qpckS8]gmaEe)D''',
        ('passacre', 'hashed.example.com'): 'abastardize abacist',
        ('passacre', 'default.example.com'): 'GEvloj!NLoJTP;$ymbp0lz<P#`3[/#4>',
    }
    extra_expected_sites = {
        'UYfDoAN9nYMdxCYtgKenzjhbc9eonu3w92ec3SAA5UbT1J3L': {
            'iterations': 10, 'schema': [[' ', 2, 'word']]
        },
    }


class SkeinYAMLTestCase(SkeinTestCaseMixin, unittest.TestCase):
    config_file = 'skein.yaml'

class SkeinSqliteTestCase(SkeinTestCaseMixin, unittest.TestCase):
    config_file = 'skein.sqlite'


def test_maybe_json():
    assert config.maybe_json('foo') == 'foo'
    assert config.maybe_json('"foo"') == 'foo'
    assert config.maybe_json('{"foo": "bar"}') == {'foo': 'bar'}

def test_no_words_file():
    # using sqlite for lazy-loading of site data, otherwise the `load` call
    # will fail too.
    c = config.load(open(os.path.join(datadir, 'no-words.sqlite'), 'rb'))
    assert c.words is None
    with pytest.raises(ValueError):
        c.generate_for_site(None, 'passacre', 'example.com')



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import uuid

from passacre._version import __version__, __sha__


signing_uuid = uuid.UUID('dd34b62f-9ed5-597e-85a2-c15d48ed6832')

__all__ = ['__version__', '__sha__', 'signing_uuid']



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

import sys
import vcversioner

# may god have mercy on my soul
from setuptools import setup


version = vcversioner.find_version(
    version_module_paths=['passacre/_version.py'],
)


with open('README.rst', 'r') as infile:
    long_description = infile.read()


extras_require = {
    'cli': ['PyYAML'],
    'clipboard': ['xerox'],
    'keccak_generation': ['cykeccak>=0.13.2'],
}

if sys.version_info > (3,):
    extras_require['skein_generation'] = ['pyskein>=0.7']

setup(
    name='passacre',
    description='better repeatable password generation',
    long_description=long_description,
    author='Aaron Gallagher',
    author_email='_@habnab.it',
    url='https://github.com/habnabit/passacre',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Security',
    ],
    license='ISC',

    version=version.version,
    packages=['passacre', 'passacre.test'],
    package_data={
        'passacre': ['schema.sql'],
        'passacre.test': ['data/*.sqlite', 'data/*.yaml', 'data/words',
                          'data/*/words', 'data/*/.passacre.*'],
    },
    extras_require=extras_require,
    entry_points={
        'console_scripts': ['passacre = passacre.application:main [cli]'],
    },
    zip_safe=False,
)



# Copyright (c) Aaron Gallagher <_@habnab.it>
# See COPYING for details.

from passacre.multibase import MultiBase

from unittest import TestCase

digits = '0123456789'
hexdigits = '0123456789abcdef'

class MultiBaseTestCase(TestCase):
    "Tests for the ``passacre.multibase.MultiBase`` class."

    def assertEncodingAndDecoding(self, mb, decoded, encoded):
        "Assert that a ``MultiBase`` encodes and decodes values correspondingly."
        self.assertEqual(mb.encode(decoded), encoded)
        self.assertEqual(mb.decode(encoded), decoded)

    def test_simple_base10(self):
        mb = MultiBase([digits, digits])

        self.assertEqual(mb.max_encodable_value, 99)

        self.assertEncodingAndDecoding(mb, 5, '05')
        self.assertEncodingAndDecoding(mb, 9, '09')
        self.assertEncodingAndDecoding(mb, 36, '36')
        self.assertEncodingAndDecoding(mb, 94, '94')

        self.assertRaises(ValueError, mb.encode, 100)
        self.assertRaises(ValueError, mb.encode, 105)

    def test_simple_base16(self):
        mb = MultiBase([hexdigits, hexdigits])

        self.assertEqual(mb.max_encodable_value, 0xff)

        self.assertEncodingAndDecoding(mb, 0x5, '05')
        self.assertEncodingAndDecoding(mb, 0xc, '0c')
        self.assertEncodingAndDecoding(mb, 0x36, '36')
        self.assertEncodingAndDecoding(mb, 0xfe, 'fe')

        self.assertRaises(ValueError, mb.encode, 0x100)
        self.assertRaises(ValueError, mb.encode, 0x105)

    def test_complex_base(self):
        mb = MultiBase(['abcd', 'abc', 'ab'])  # 4 * 3 * 2 == 24

        self.assertEqual(mb.max_encodable_value, 23)

        self.assertEncodingAndDecoding(mb, 0, 'aaa')
        self.assertEncodingAndDecoding(mb, 5, 'acb')
        self.assertEncodingAndDecoding(mb, 9, 'bbb')
        self.assertEncodingAndDecoding(mb, 11, 'bcb')
        self.assertEncodingAndDecoding(mb, 17, 'ccb')
        self.assertEncodingAndDecoding(mb, 23, 'dcb')

        self.assertRaises(ValueError, mb.encode, 24)

    def test_decoding_exceptions(self):
        mb = MultiBase([digits, digits])

        self.assertRaises(ValueError, mb.decode, '999')
        self.assertRaises(ValueError, mb.decode, '9')
        self.assertRaises(ValueError, mb.decode, '9a')
